from __future__ import annotations

import asyncio
import sys
import traceback
from time import sleep
from typing import Optional

import clr
import PalmSens
from PalmSens import Method, MuxModel
from PalmSens.Comm import CommManager, MuxType
from PalmSens.Plottables import (
    Curve,
    CurveEventHandler,
    EISData,
    EISDataEventHandler,
)
from System import EventHandler  # type: ignore

from .._data._shared import ArrayType, _get_values_from_NETArray
from .._methods import CURRENT_RANGE, BaseTechnique
from ..data import Measurement
from ._common import Callback, Instrument, create_future, firmware_warning

WINDOWS = sys.platform == 'win32'
LINUX = not WINDOWS

if WINDOWS:
    from PalmSens.Windows.Devices import (
        BLEDevice,
        BluetoothDevice,
        FTDIDevice,
        USBCDCDevice,
        WinUSBDevice,
    )
else:
    from PalmSens.Core.Linux.Comm.Devices import FTDIDevice, SerialPortDevice


def discover(
    ftdi: bool = False,
    usbcdc: bool = True,
    winusb: bool = True,
    bluetooth: bool = False,
    serial: bool = True,
) -> list[Instrument]:
    """Discover instruments.

    Parameters
    ----------
    ftdi : bool
        If True, discover ftdi devices
    usbcdc : bool
        If True, discover usbcdc devices (Windows only)
    winusb : bool
        If True, discover winusb devices (Windows only)
    bluetooth : bool
        If True, discover bluetooth devices (Windows only)
    serial : bool
        If True, discover serial devices
    """
    available_instruments = []
    args = [''] if WINDOWS else []

    if ftdi:
        ftdi_instruments = FTDIDevice.DiscoverDevices(*args)
        for ftdi_instrument in ftdi_instruments[0]:
            available_instruments.append(
                Instrument(
                    id=ftdi_instrument.ToString(),
                    interface='ftdi',
                    device=ftdi_instrument,
                )
            )

    if WINDOWS and usbcdc:
        usbcdc_instruments = USBCDCDevice.DiscoverDevices(*args)
        for usbcdc_instrument in usbcdc_instruments[0]:
            available_instruments.append(
                Instrument(
                    id=usbcdc_instrument.ToString(),
                    interface='usbcdc',
                    device=usbcdc_instrument,
                )
            )

    if WINDOWS and winusb:
        winusb_instruments = WinUSBDevice.DiscoverDevices(*args)
        for winusb_instrument in winusb_instruments[0]:
            available_instruments.append(
                Instrument(
                    id=winusb_instrument.ToString(),
                    interface='winusb',
                    device=winusb_instrument,
                )
            )

    if WINDOWS and bluetooth:
        ble_instruments = BLEDevice.DiscoverDevices(*args)
        for ble_instrument in ble_instruments[0]:
            available_instruments.append(
                Instrument(
                    id=ble_instrument.ToString(),
                    interface='ble',
                    device=ble_instrument,
                )
            )

        bluetooth_instruments = BluetoothDevice.DiscoverDevices(*args)
        for bluetooth_instrument in bluetooth_instruments[0]:
            available_instruments.append(
                Instrument(
                    id=bluetooth_instrument.ToString(),
                    interface='bluetooth',
                    device=bluetooth_instrument,
                )
            )

    if LINUX and serial:
        serial_instruments = SerialPortDevice.DiscoverDevices(*args)
        for serial_instrument in serial_instruments:
            available_instruments.append(
                Instrument(
                    id=serial_instrument.ToString(),
                    interface='serial',
                    device=serial_instrument,
                )
            )

    available_instruments.sort(key=lambda instrument: instrument.id)

    return available_instruments


def connect(
    instrument: Optional[Instrument] = None,
) -> InstrumentManager:
    """Connect to instrument and return InstrumentManager.

    Parameters
    ----------
    instrument : Instrument, optional
        Connect to this instrument.
        If not specified, automatically discover and connect to the first instrument.

    Returns
    -------
    manager : InstrumentManager
        Return instance of `InstrumentManager` connected to the given instrument.
        The connection will be terminated after the context ends.
    """
    # connect to first device if not specified
    if not instrument:
        available_instruments = discover()

        if not available_instruments:
            raise ConnectionError('No instruments were discovered.')

        # connect to first instrument
        instrument = available_instruments[0]

    manager = InstrumentManager(instrument)
    manager.connect()
    return manager


class InstrumentManager:
    """Instrument manager for PalmSens instruments.

    Parameters
    ----------
    instrument: Instrument
        Instrument to connect to, use `discover()` to find connected instruments.
    callback: Callback, optional
        If specified, call this function on every new set of data points.
        New data points are batched, and contain all points since the last
        time it was called. Each point is a dictionary containing
        `frequency`, `z_re`, `z_im` for impedimetric techniques and
        `index`, `x`, `x_unit`, `x_type`, `y`, `y_unit` and `y_type` for
        non-impedimetric techniques.
    """

    def __init__(self, instrument: Instrument, *, callback: Optional[Callback] = None):
        self.callback = callback
        """This callback is called on every data point."""

        self.instrument = instrument
        """Instrument to connect to."""

        self.__comm = None
        self.__measuring = False
        self.__active_measurement = None
        self.__active_measurement_error = None

    def __repr__(self):
        return (
            f'{self.__class__.__name__}({self.instrument.id}, connected={self.is_connected()})'
        )

    def __enter__(self):
        if not self.is_connected():
            self.connect()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.disconnect()

    def is_connected(self) -> bool:
        """Return True if an instrument connection exists."""
        return self.__comm is not None

    def connect(self):
        """Connect to instrument."""
        if self.__comm is not None:
            print(
                'An instance of the InstrumentManager can only be connected to one instrument at a time'
            )
            return 0
        try:
            __instrument = self.instrument.device
            __instrument.Open()
            self.__comm = CommManager(__instrument)

            firmware_warning(self.__comm.Capabilities)

            return 1
        except Exception:
            traceback.print_exc()
            try:
                __instrument.Close()
            except Exception:
                pass
            return 0

    def set_cell(self, cell_on: bool):
        """Turn the cell on or off.

        Parameters
        ----------
        cell_on : bool
            If true, turn on the cell
        """
        if self.__comm is None:
            print('Not connected to an instrument')
            return 0

        self.__comm.ClientConnection.Semaphore.Wait()

        try:
            self.__comm.CellOn = cell_on
            self.__comm.ClientConnection.Semaphore.Release()
        except Exception:
            traceback.print_exc()

            if self.__comm.ClientConnection.Semaphore.CurrentCount == 0:
                # release lock on library (required when communicating with instrument)
                self.__comm.ClientConnection.Semaphore.Release()

    def set_potential(self, potential: float):
        """Set the potential of the cell.

        Parameters
        ----------
        potential : float
            Potential in V
        """
        if self.__comm is None:
            print('Not connected to an instrument')
            return 0

        self.__comm.ClientConnection.Semaphore.Wait()

        try:
            self.__comm.Potential = potential
            self.__comm.ClientConnection.Semaphore.Release()
        except Exception:
            traceback.print_exc()

            if self.__comm.ClientConnection.Semaphore.CurrentCount == 0:
                # release lock on library (required when communicating with instrument)
                self.__comm.ClientConnection.Semaphore.Release()

    def set_current_range(self, current_range: CURRENT_RANGE):
        """Set the current range for the cell.

        Parameters
        ----------
        current_range: CURRENT_RANGE
            Set the current range, use `pypalmsens.settings.CURRENT_RANGE`.
        """
        if self.__comm is None:
            print('Not connected to an instrument')
            return 0

        self.__comm.ClientConnection.Semaphore.Wait()

        try:
            self.__comm.CurrentRange = current_range._to_psobj()
            self.__comm.ClientConnection.Semaphore.Release()
        except Exception:
            traceback.print_exc()

            if self.__comm.ClientConnection.Semaphore.CurrentCount == 0:
                # release lock on library (required when communicating with instrument)
                self.__comm.ClientConnection.Semaphore.Release()

    def read_current(self) -> float:
        """Read the current in µA.

        Returns
        -------
        float
            Current in µA."
        """
        if self.__comm is None:
            raise ConnectionError('Not connected to an instrument')

        self.__comm.ClientConnection.Semaphore.Wait()

        try:
            current = self.__comm.Current  # in µA
            self.__comm.ClientConnection.Semaphore.Release()
            return current
        except Exception:
            traceback.print_exc()

            if self.__comm.ClientConnection.Semaphore.CurrentCount == 0:
                # release lock on library (required when communicating with instrument)
                self.__comm.ClientConnection.Semaphore.Release()

    def read_potential(self) -> float:
        """Read the potential in V.

        Returns
        -------
        float
            Potential in V."""

        if self.__comm is None:
            raise ConnectionError('Not connected to an instrument')

        self.__comm.ClientConnection.Semaphore.Wait()

        try:
            potential = self.__comm.Potential  # in V
            self.__comm.ClientConnection.Semaphore.Release()
            return potential
        except Exception:
            traceback.print_exc()

            if self.__comm.ClientConnection.Semaphore.CurrentCount == 0:
                # release lock on library (required when communicating with instrument)
                self.__comm.ClientConnection.Semaphore.Release()

    def get_instrument_serial(self) -> str:
        """Return instrument serial number.

        Returns
        -------
        str
            Instrument serial.
        """
        if self.__comm is None:
            raise ConnectionError('Not connected to an instrument')

        self.__comm.ClientConnection.Semaphore.Wait()

        try:
            serial = self.__comm.DeviceSerial.ToString()
            self.__comm.ClientConnection.Semaphore.Release()
            return serial
        except Exception:
            traceback.print_exc()

            if self.__comm.ClientConnection.Semaphore.CurrentCount == 0:
                # release lock on library (required when communicating with instrument)
                self.__comm.ClientConnection.Semaphore.Release()

    def validate_method(self, psmethod):
        """Validate method."""
        if self.__comm is None:
            print('Not connected to an instrument')
            return False, None

        errors = psmethod.Validate(self.__comm.Capabilities)

        if any(error.IsFatal for error in errors):
            return False, 'Method not compatible:\n' + '\n'.join(
                [error.Message for error in errors]
            )

        return True, None

    def measure(self, method: BaseTechnique):
        """Start measurement using given method parameters.

        Parameters
        ----------
        method: MethodParameters
            Method parameters for measurement
        """
        psmethod = method._to_psmethod()
        if self.__comm is None:
            print('Not connected to an instrument')
            return None

        self.__active_measurement_error = None

        is_valid, message = self.validate_method(psmethod)
        if not is_valid:
            raise ValueError(message)

        loop = asyncio.new_event_loop()
        begin_measurement_event = asyncio.Event()
        end_measurement_event = asyncio.Event()

        def begin_measurement(measurement):
            self.__active_measurement = measurement
            begin_measurement_event.set()

        def end_measurement():
            self.__measuring = False
            end_measurement_event.set()

        def curve_new_data_added(curve, start, count):
            data = []
            for i in range(start, start + count):
                point = {}
                point['index'] = i + 1
                point['x'] = _get_values_from_NETArray(curve.XAxisDataArray, start=i, count=1)[
                    0
                ]
                point['x_unit'] = curve.XUnit.ToString()
                point['x_type'] = ArrayType(curve.XAxisDataArray.ArrayType).name
                point['y'] = _get_values_from_NETArray(curve.YAxisDataArray, start=i, count=1)[
                    0
                ]
                point['y_unit'] = curve.YUnit.ToString()
                point['y_type'] = ArrayType(curve.YAxisDataArray.ArrayType).name
                data.append(point)
            self.callback(data)

        def eis_data_new_data_added(eis_data, start, count):
            data = []
            arrays = eis_data.EISDataSet.GetDataArrays()
            for i in range(start, start + count):
                point = {}
                point['index'] = i + 1
                for array in arrays:
                    array_type = ArrayType(array.ArrayType)
                    if array_type == ArrayType.Frequency:
                        point['frequency'] = _get_values_from_NETArray(array, start=i, count=1)[
                            0
                        ]
                    elif array_type == ArrayType.ZRe:
                        point['zre'] = _get_values_from_NETArray(array, start=i, count=1)[0]
                    elif array_type == ArrayType.ZIm:
                        point['zim'] = _get_values_from_NETArray(array, start=i, count=1)[0]
                data.append(point)
            self.callback(data)

        def comm_error():
            self.__measuring = False
            self.__active_measurement_error = (
                'measurement failed due to a communication or parsing error'
            )
            begin_measurement_event.set()
            end_measurement_event.set()

        def begin_measurement_callback(sender, measurement):
            loop.call_soon_threadsafe(lambda: begin_measurement(measurement))

        def end_measurement_callback(sender, args):
            loop.call_soon_threadsafe(end_measurement)

        def curve_data_added_callback(curve, args):
            start = args.StartIndex
            count = curve.NPoints - start
            loop.call_soon_threadsafe(lambda: curve_new_data_added(curve, start, count))

        def curve_finished_callback(curve, args):
            curve.NewDataAdded -= curve_data_added_handler
            curve.Finished -= curve_finished_handler

        def begin_receive_curve_callback(sender, args):
            curve = args.GetCurve()
            curve.NewDataAdded += curve_data_added_handler
            curve.Finished += curve_finished_handler

        def eis_data_data_added_callback(eis_data, args):
            start = args.Index
            count = 1
            loop.call_soon_threadsafe(lambda: eis_data_new_data_added(eis_data, start, count))

        def eis_data_finished_callback(eis_data, args):
            eis_data.NewDataAdded -= eis_data_data_added_handler
            eis_data.Finished -= eis_data_finished_handler

        def begin_receive_eis_data_callback(sender, eis_data):
            eis_data.NewDataAdded += eis_data_data_added_handler
            eis_data.Finished += eis_data_finished_handler

        def comm_error_callback(sender, args):
            loop.call_soon_threadsafe(comm_error)

        begin_measurement_handler = CommManager.BeginMeasurementEventHandler(
            begin_measurement_callback
        )
        end_measurement_handler = EventHandler(end_measurement_callback)
        begin_receive_curve_handler = CurveEventHandler(begin_receive_curve_callback)
        curve_data_added_handler = Curve.NewDataAddedEventHandler(curve_data_added_callback)
        curve_finished_handler = EventHandler(curve_finished_callback)
        eis_data_finished_handler = EventHandler(eis_data_finished_callback)
        begin_receive_eis_data_handler = EISDataEventHandler(begin_receive_eis_data_callback)
        eis_data_data_added_handler = EISData.NewDataEventHandler(eis_data_data_added_callback)
        comm_error_handler = EventHandler(comm_error_callback)

        try:
            # subscribe to events indicating the start and end of the measurement
            self.__comm.BeginMeasurement += begin_measurement_handler
            self.__comm.EndMeasurement += end_measurement_handler
            self.__comm.Disconnected += comm_error_handler

            if self.callback is not None:
                self.__comm.BeginReceiveEISData += begin_receive_eis_data_handler
                self.__comm.BeginReceiveCurve += begin_receive_curve_handler

            async def await_measurement():
                # obtain lock on library (required when communicating with instrument)
                await create_future(self.__comm.ClientConnection.Semaphore.WaitAsync())

                # send and execute the method on the instrument
                _ = self.__comm.Measure(psmethod)
                self.__measuring = True

                # release lock on library (required when communicating with instrument)
                self.__comm.ClientConnection.Semaphore.Release()

                await begin_measurement_event.wait()
                await end_measurement_event.wait()

            loop.run_until_complete(await_measurement())
            loop.close()

            # unsubscribe to events indicating the start and end of the measurement
            self.__comm.BeginMeasurement -= begin_measurement_handler
            self.__comm.EndMeasurement -= end_measurement_handler
            self.__comm.Disconnected -= comm_error_handler

            if self.callback is not None:
                self.__comm.BeginReceiveEISData -= begin_receive_eis_data_handler
                self.__comm.BeginReceiveCurve -= begin_receive_curve_handler

            if self.__active_measurement_error is not None:
                print(self.__active_measurement_error)
                return None

            measurement = self.__active_measurement
            self.__active_measurement = None
            return Measurement(psmeasurement=measurement)

        except Exception:
            traceback.print_exc()

            if self.__comm.ClientConnection.Semaphore.CurrentCount == 0:
                # release lock on library (required when communicating with instrument)
                self.__comm.ClientConnection.Semaphore.Release()

            self.__active_measurement = None
            self.__comm.BeginMeasurement -= begin_measurement_handler
            self.__comm.EndMeasurement -= end_measurement_handler
            self.__comm.Disconnected -= comm_error_handler

            if self.callback is not None:
                self.__comm.BeginReceiveEISData -= begin_receive_eis_data_handler
                self.__comm.BeginReceiveCurve -= begin_receive_curve_handler

            self.__measuring = False
            return None

    def wait_digital_trigger(self, wait_for_high):
        """Wait for digital trigger.

        Parameters
        ----------
        wait_for_high: ...
            ...
        """
        if self.__comm is None:
            print('Not connected to an instrument')
            return 0

        try:
            # obtain lock on library (required when communicating with instrument)
            self.__comm.ClientConnection.Semaphore.Wait()

            while True:
                if self.__comm.DigitalLineD0 == wait_for_high:
                    break
                sleep(0.05)

            self.__comm.ClientConnection.Semaphore.Release()

        except Exception:
            traceback.print_exc()

            if self.__comm.ClientConnection.Semaphore.CurrentCount == 0:
                # release lock on library (required when communicating with instrument)
                self.__comm.ClientConnection.Semaphore.Release()

    def abort(self):
        """Abort measurement."""
        if self.__comm is None:
            print('Not connected to an instrument')
            return 0

        if self.__measuring is False:
            return 0

        self.__comm.ClientConnection.Semaphore.Wait()
        try:
            self.__comm.Abort()
        except Exception:
            traceback.print_exc()

            if self.__comm.ClientConnection.Semaphore.CurrentCount == 0:
                # release lock on library (required when communicating with instrument)
                self.__comm.ClientConnection.Semaphore.Release()

    def initialize_multiplexer(self, mux_model: int) -> int:
        """Initialize the multiplexer.

        Parameters
        ----------
        mux_model: int
            The model of the multiplexer. 0 = 8 channel, 1 = 16 channel, 2 = 32 channel.

        Returns
        -------
        int
            Number of available multiplexes channels
        """
        if self.__comm is None:
            print('Not connected to an instrument')
            return 0

        self.__comm.ClientConnection.Semaphore.Wait()

        try:
            model = MuxModel(mux_model)

            if model == MuxModel.MUX8R2 and (
                self.__comm.ClientConnection.GetType().Equals(
                    clr.GetClrType(PalmSens.Comm.ClientConnectionPS4)
                )
                or self.__comm.ClientConnection.GetType().Equals(
                    clr.GetClrType(PalmSens.Comm.ClientConnectionMS)
                )
            ):
                self.__comm.ClientConnection.ReadMuxInfo()

            self.__comm.Capabilities.MuxModel = model

            if self.__comm.Capabilities.MuxModel == MuxModel.MUX8:
                self.__comm.Capabilities.NumMuxChannels = 8
            elif self.__comm.Capabilities.MuxModel == MuxModel.MUX16:
                self.__comm.Capabilities.NumMuxChannels = 16
            elif self.__comm.Capabilities.MuxModel == MuxModel.MUX8R2:
                self.__comm.ClientConnection.ReadMuxInfo()

            self.__comm.ClientConnection.Semaphore.Release()

            return self.__comm.Capabilities.NumMuxChannels
        except Exception:
            traceback.print_exc()

            if self.__comm.ClientConnection.Semaphore.CurrentCount == 0:
                # release lock on library (required when communicating with instrument)
                self.__comm.ClientConnection.Semaphore.Release()

            raise Exception(
                'Failed to read MUX info. Please check the connection, restart the instrument and try again.'
            )

    def set_mux8r2_settings(
        self,
        connect_sense_to_working_electrode: bool = False,
        combine_reference_and_counter_electrodes: bool = False,
        use_channel_1_reference_and_counter_electrodes: bool = False,
        set_unselected_channel_working_electrode: int = 0,
    ):
        """Set the settings for the Mux8R2 multiplexer.

        Parameters
        ---------
        connect_sense_to_working_electrode: float
            Connect the sense electrode to the working electrode. Default is False.
        combine_reference_and_counter_electrodes: float
            Combine the reference and counter electrodes. Default is False.
        use_channel_1_reference_and_counter_electrodes: float
            Use channel 1 reference and counter electrodes for all working electrodes. Default is False.
        set_unselected_channel_working_electrode: float
            Set the unselected channel working electrode to disconnected/floating (0), ground (1), or standby potential (2). Default is 0.
        """
        if self.__comm is None:
            print('Not connected to an instrument')
            return

        if self.__comm.Capabilities.MuxModel != MuxModel.MUX8R2:
            return

        mux_settings = Method.MuxSettings(False)
        mux_settings.ConnSEWE = connect_sense_to_working_electrode
        mux_settings.ConnectCERE = combine_reference_and_counter_electrodes
        mux_settings.CommonCERE = use_channel_1_reference_and_counter_electrodes
        mux_settings.UnselWE = Method.MuxSettings.UnselWESetting(
            set_unselected_channel_working_electrode
        )

        self.__comm.ClientConnection.Semaphore.Wait()

        try:
            self.__comm.ClientConnection.SetMuxSettings(MuxType(1), mux_settings)
            self.__comm.ClientConnection.Semaphore.Release()
        except Exception:
            traceback.print_exc()

            if self.__comm.ClientConnection.Semaphore.CurrentCount == 0:
                # release lock on library (required when communicating with instrument)
                self.__comm.ClientConnection.Semaphore.Release()

    def set_multiplexer_channel(self, channel: int):
        """Sets the multiplexer channel.

        Parameters
        ----------
        channel : int
            Index of the channel to set.
        """
        if self.__comm is None:
            print('Not connected to an instrument')
            return 0

        self.__comm.ClientConnection.Semaphore.Wait()

        try:
            self.__comm.ClientConnection.SetMuxChannel(channel)
            self.__comm.ClientConnection.Semaphore.Release()
        except Exception:
            traceback.print_exc()

            if self.__comm.ClientConnection.Semaphore.CurrentCount == 0:
                # release lock on library (required when communicating with instrument)
                self.__comm.ClientConnection.Semaphore.Release()

    def disconnect(self):
        """Disconnect from the instrument."""
        if self.__comm is None:
            return 0
        try:
            self.__comm.Disconnect()
            self.__comm = None
            self.__measuring = False
            return 1
        except Exception:
            traceback.print_exc()
            return 0



from __future__ import annotations

from abc import abstractmethod
from typing import Any, ClassVar, Protocol, Type, runtime_checkable

import attrs
from PalmSens import Method as PSMethod


@runtime_checkable
class BaseSettings(Protocol):
    """Protocol to provide generic methods for parameters."""

    @abstractmethod
    def _update_psmethod(self, psmethod: PSMethod, /): ...

    @abstractmethod
    def _update_params(self, psmethod: PSMethod, /): ...


@runtime_checkable
class BaseTechnique(Protocol):
    """Protocol to provide base methods for method classes."""

    __attrs_attrs__: ClassVar[list[attrs.Attribute]] = []
    _id: str
    _registry: dict[str, Type[BaseTechnique]] = {}

    def __init_subclass__(cls, **kwargs) -> None:
        super().__init_subclass__(**kwargs)
        cls._registry[cls._id] = cls

    def to_dict(self) -> dict[str, Any]:
        """Return the technique instance as a new key/value dictionary mapping."""
        return attrs.asdict(self, filter=lambda a, _: not a.name.startswith('_'))

    @classmethod
    def from_method_id(cls, id: str) -> BaseTechnique:
        """Create new instance of appropriate technique from method ID."""
        new = cls._registry[id]
        return new()

    @classmethod
    def _from_psmethod(cls, psmethod: PSMethod, /) -> BaseTechnique:
        """Generate parameters from dotnet method object."""
        new = cls.from_method_id(psmethod.MethodID)
        new._update_params(psmethod)
        new._update_params_nested(psmethod)
        return new

    @abstractmethod
    def _update_params(self, psmethod: PSMethod, /) -> None: ...

    def _update_params_nested(self, psmethod: PSMethod, /) -> None:
        """Retrieve and convert dotnet method for nested field parameters."""
        for field in self.__attrs_attrs__:
            attribute = getattr(self, field.name)
            try:
                # Update parameters if attribute has the `update_params` method
                attribute._update_params(psmethod)
            except AttributeError:
                pass

    def _to_psmethod(self) -> PSMethod:
        """Convert parameters to dotnet method."""
        psmethod = PSMethod.FromMethodID(self._id)

        self._update_psmethod(psmethod)
        self._update_psmethod_nested(psmethod)
        return psmethod

    @abstractmethod
    def _update_psmethod(self, psmethod: PSMethod, /) -> None: ...

    def _update_psmethod_nested(self, psmethod: PSMethod, /) -> None:
        """Convert and set field parameters on dotnet method."""
        for field in self.__attrs_attrs__:
            attribute = getattr(self, field.name)
            try:
                # Update parameters if attribute has the `update_params` method
                attribute._update_psmethod(psmethod)
            except AttributeError:
                pass



from __future__ import annotations

import sys

# Load dotnet platform dependencies and init SDK
if sys.platform == 'win32':
    from ._lib.windows import sdk_version
else:
    from ._lib.mono import sdk_version

__version__ = '1.2.2'
__sdk_version__ = sdk_version

from . import data, fitting, mixed_mode, settings
from ._instruments.instrument_manager import (
    InstrumentManager,
    connect,
    discover,
)
from ._instruments.instrument_manager_async import (
    InstrumentManagerAsync,
    connect_async,
    discover_async,
)
from ._instruments.instrument_pool import InstrumentPool
from ._instruments.instrument_pool_async import InstrumentPoolAsync
from ._io import load_method_file, load_session_file, save_method_file, save_session_file
from ._methods.techniques import (
    ACVoltammetry,
    ChronoAmperometry,
    ChronoCoulometry,
    ChronoPotentiometry,
    CyclicVoltammetry,
    DifferentialPulseVoltammetry,
    ElectrochemicalImpedanceSpectroscopy,
    FastAmperometry,
    FastCyclicVoltammetry,
    FastGalvanostaticImpedanceSpectroscopy,
    FastImpedanceSpectroscopy,
    GalvanostaticImpedanceSpectroscopy,
    LinearSweepPotentiometry,
    LinearSweepVoltammetry,
    MethodScript,
    MultiStepAmperometry,
    MultiStepPotentiometry,
    NormalPulseVoltammetry,
    OpenCircuitPotentiometry,
    PulsedAmperometricDetection,
    SquareWaveVoltammetry,
    StrippingChronoPotentiometry,
)

__all__ = [
    'settings',
    'data',
    'fitting',
    'mixed_mode',
    'connect',
    'connect_async',
    'discover',
    'discover_async',
    'load_method_file',
    'load_session_file',
    'save_method_file',
    'save_session_file',
    'InstrumentManager',
    'FastAmperometry',
    'InstrumentManagerAsync',
    'InstrumentPool',
    'InstrumentPoolAsync',
    'ACVoltammetry',
    'ChronoAmperometry',
    'ChronoCoulometry',
    'ChronoPotentiometry',
    'CyclicVoltammetry',
    'DifferentialPulseVoltammetry',
    'ElectrochemicalImpedanceSpectroscopy',
    'FastCyclicVoltammetry',
    'FastGalvanostaticImpedanceSpectroscopy',
    'FastImpedanceSpectroscopy',
    'GalvanostaticImpedanceSpectroscopy',
    'LinearSweepPotentiometry',
    'LinearSweepVoltammetry',
    'MethodScript',
    'MultiStepAmperometry',
    'MultiStepPotentiometry',
    'NormalPulseVoltammetry',
    'OpenCircuitPotentiometry',
    'PulsedAmperometricDetection',
    'SquareWaveVoltammetry',
    'StrippingChronoPotentiometry',
]



from __future__ import annotations

from ._shared import CURRENT_RANGE, POTENTIAL_RANGE
from .base import BaseTechnique
from .method import Method
from .settings import BaseSettings

__all__ = [
    'Method',
    'BaseTechnique',
    'BaseSettings',
    'CURRENT_RANGE',
    'POTENTIAL_RANGE',
]



from __future__ import annotations

from ._methods.mixed_mode import (
    ConstantE,
    ConstantI,
    Impedance,
    MixedMode,
    OpenCircuit,
    SweepE,
)

__all__ = [
    'MixedMode',
    'ConstantE',
    'ConstantI',
    'Impedance',
    'OpenCircuit',
    'SweepE',
]



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Callable, Sequence

from .._methods import BaseTechnique
from ._common import Instrument
from .instrument_manager_async import InstrumentManagerAsync
from .instrument_pool_async import InstrumentPoolAsync

if TYPE_CHECKING:
    from .._data.measurement import Measurement


class InstrumentPool:
    """Manages a set of instrument.

    Most calls are run asynchronously in the background,
    which means that measurements are running in parallel.

    This is a thin wrapper around the `InstrumentManagerAsync`.


    Parameters
    ----------
    devices_or_managers : list[Instrument | InstrumentManagerAsync]
        List of devices or managers.
    callback : Callable, optional
        Optional callable to set on instrument managers
    """

    def __init__(
        self,
        devices_or_managers: Sequence[Instrument | InstrumentManagerAsync],
        *,
        callback: None | Callable = None,
    ):
        self._async = InstrumentPoolAsync(devices_or_managers, callback=callback)
        self._loop = asyncio.new_event_loop()

        self.managers = self._async.managers
        """List of instruments managers in the pool."""

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        return self.disconnect()

    def __iter__(self):
        yield from self.managers

    def connect(self) -> None:
        """Connect all instrument managers in the pool."""
        self._loop.run_until_complete(self._async.connect())

    def disconnect(self) -> None:
        """Disconnect all instrument managers in the pool."""
        self._loop.run_until_complete(self._async.disconnect())

    def is_connected(self) -> bool:
        """Return true if all managers in the pool are connected."""
        return self._async.is_connected()

    def is_disconnected(self) -> bool:
        """Return true if all managers in the pool are disconnected."""
        return self._async.is_disconnected()

    def remove(self, manager: InstrumentManagerAsync) -> None:
        """Close and remove manager from pool.

        Parameters
        ----------
        manager : InstrumentManagerAsync
            Instance of an instrument manager.
        """
        self._loop.run_until_complete(self._async.remove(manager))

    def add(self, manager: InstrumentManagerAsync) -> None:
        """Open and add manager to the pool.

        Parameters
        ----------
        manager : InstrumentManagerAsync
            Instance of an instrument manager.
        """
        self._loop.run_until_complete(self._async.add(manager))

    def measure(self, method: BaseTechnique) -> list[Measurement]:
        """Concurrently run measurement on all managers in the pool.

        For hardware synchronization, set `use_hardware_sync` on the method.
        In addition, the pool must contain:
        - channels from a single multi-channel instrument only
        - the first channel of the multi-channel instrument
        - at least two channels

        All instruments are prepared and put in a waiting state.
        The measurements are started via a hardware sync trigger on channel 1.

        Parameters
        ----------
        method : MethodSettings
            Method parameters for measurement.
        """
        return self._loop.run_until_complete(self._async.measure(method=method))



from __future__ import annotations

from enum import Enum
from typing import Optional


class ArrayType(Enum):
    """Data array type for standard arrays."""

    Unspecified = -1
    """Unspecified"""
    Time = 0
    """Time / s"""
    Potential = 1
    """Potential / V"""
    Current = 2
    """Current / μA"""
    Charge = 3
    """Charge"""
    ExtraValue = 4
    """ExtraValue"""
    Frequency = 5
    """Frequency"""
    Phase = 6
    """Phase"""
    ZRe = 7
    """Z real"""
    ZIm = 8
    """Z imaginary"""
    Iac = 9
    """I AC values"""
    Z = 10
    """Z"""
    Y = 11
    """Y"""
    YRe = 12
    """Y real"""
    YIm = 13
    """Y imaginary"""
    Cs = 14
    """Cs"""
    CsRe = 15
    """Cs real"""
    CsIm = 16
    """Cs imaginary"""
    Index = 17
    """Index"""
    Admittance = 18
    """Admittance"""
    Concentration = 19
    """Concentration"""
    Signal = 20
    """Signal"""
    Func = 21
    """Func"""
    Integral = 22
    """Integral"""
    AuxInput = 23
    """Auxillary input"""
    BipotCurrent = 24
    """Bipot current"""
    BipotPotential = 25
    """Bipot potential"""
    ReverseCurrent = 26
    """Reverse current"""
    CEPotential = 27
    """CE potential"""
    DCCurrent = 28
    """DC current"""
    ForwardCurrent = 29
    """Forward current"""
    PotentialExtraRE = 30
    """Potential setpoint measured back on RE"""
    CurrentExtraWE = 31
    """Current setpoint measured back on WE"""
    InverseDerative_dtdE = 32
    """Inverse derivative dt/dE"""
    mEdc = 33
    """Measured applied DC"""
    Eac = 34
    """E AC values"""
    MeasuredStepStartIndex = 35
    """MeasuredStepStartIndex"""
    miDC = 36
    """Measured I DC values"""
    SE2vsXPotential = 37
    """SE2 vs XPotential"""

    @classmethod
    def _missing_(cls, value):
        return cls.Unspecified


class Status(Enum):
    Unknown = -1
    OK = 0
    Overload = 1
    Underload = 2


def _get_values_from_NETArray(array, start: int = 0, count: Optional[int] = None):
    if not count:
        count = array.Count

    values = []
    for i in range(start, start + count):
        value = array.get_Item(i)
        values.append(float(value.Value))
    return values


def __get_currentranges_from_currentarray(
    arraycurrents, start: int = 0, count: Optional[int] = None
):
    if not count:
        count = arraycurrents.Count
    values = []
    if ArrayType(arraycurrents.ArrayType) == ArrayType.Current:
        for i in range(start, count):
            value = arraycurrents.get_Item(i)
            values.append(str(value.CurrentRange.ToString()))
    return values


def __get_status_from_current_or_potentialarray(
    array, start: int = 0, count: Optional[int] = None
):
    if not count:
        count = array.Count

    values = []
    for i in range(start, count):
        value = array.get_Item(i)
        values.append(str(Status(value.ReadingStatus)))
    return values



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any, Union

from attrs import asdict
from PalmSens import Method as PSMethod
from PalmSens.Devices import PalmSens4Capabilities

from . import techniques

if TYPE_CHECKING:
    from .techniques import BaseTechnique


class Method:
    """Wrapper for PalmSens.Method."""

    def __init__(self, *, psmethod: PSMethod):
        self.psmethod = psmethod

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(name={self.name!r}, id={self.id!r})'

    @property
    def id(self) -> str:
        """Unique id for method."""
        return self.psmethod.MethodID

    @property
    def name(self) -> str:
        """Name for the technique."""
        return self.psmethod.Name

    @property
    def short_name(self) -> str:
        """Short name for the technique."""
        return self.psmethod.ShortName

    @property
    def filename(self) -> Union[Path, None]:
        """Filename for the method if applicable."""
        fn = self.psmethod.MethodFilename
        if fn:
            return Path(fn)
        return None

    def get_estimated_duration(self, *, instrument_manager=None):
        """Get the estimated duration for this method.

        Parameters
        ----------
        instrument_manager : InstrumentManager
            Specifies the instrument manager to get the connected instruments capabilities from,
            If not specified it will use the PalmSens4 capabilities to determine the estimated duration.
        """
        if instrument_manager is None or instrument_manager.__comm is None:
            instrument_capabilities = PalmSens4Capabilities()
        else:
            instrument_capabilities = instrument_manager.__comm.Capabilities
        return self.psmethod.GetMinimumEstimatedMeasurementDuration(instrument_capabilities)

    @property
    def technique_number(self) -> int:
        """The technique number used in the firmware."""
        return self.psmethod.Technique

    def to_settings(self) -> BaseTechnique:
        """Extract techniques parameters as dataclass."""
        return techniques.BaseTechnique._from_psmethod(self.psmethod)

    def to_dict(self) -> dict[str, Any]:
        """Return dictionary with technique parameters."""
        return asdict(self.to_settings())



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Any, Awaitable, Callable, Sequence

from .._methods import BaseTechnique
from ._common import Instrument
from .instrument_manager_async import InstrumentManagerAsync

if TYPE_CHECKING:
    from .._data.measurement import Measurement


class InstrumentPoolAsync:
    """Manages a set of instrument.

    Most calls are run asynchronously in the background,
    which means that measurements are running in parallel.

    Parameters
    ----------
    devices_or_managers : list[Instrument | InstrumentManagerAsync]
        List of devices or managers.
    callback : Callable, optional
        Optional callable to set on instrument managers
    """

    def __init__(
        self,
        devices_or_managers: Sequence[Instrument | InstrumentManagerAsync],
        *,
        callback: None | Callable = None,
    ):
        self.managers = []
        """List of instruments managers in the pool."""

        for item in devices_or_managers:
            if isinstance(item, Instrument):
                self.managers.append(InstrumentManagerAsync(item, callback=callback))
            else:
                if callback:
                    item.callback = callback
                self.managers.append(item)

    async def __aenter__(self):
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        return await self.disconnect()

    def __iter__(self):
        yield from self.managers

    async def connect(self) -> None:
        """Connect all instrument managers in the pool."""
        tasks = []
        for manager in self.managers:
            tasks.append(manager.connect())
        await asyncio.gather(*tasks)

    async def disconnect(self) -> None:
        """Disconnect all instrument managers in the pool."""
        tasks = []
        for manager in self.managers:
            tasks.append(manager.disconnect())
        await asyncio.gather(*tasks)

    def is_connected(self) -> bool:
        """Return true if all managers in the pool are connected."""
        return all(manager.is_connected() for manager in self.managers)

    def is_disconnected(self) -> bool:
        """Return true if all managers in the pool are disconnected."""
        return not any(manager.is_connected() for manager in self.managers)

    async def remove(self, manager: InstrumentManagerAsync) -> None:
        """Close and remove manager from pool.

        Parameters
        ----------
        manager : InstrumentManagerAsync
            Instance of an instrument manager.
        """
        self.managers.remove(manager)
        await manager.disconnect()

    async def add(self, manager: InstrumentManagerAsync) -> None:
        """Open and add manager to the pool.

        Parameters
        ----------
        manager : InstrumentManagerAsync
            Instance of an instrument manager.
        """
        await manager.connect()
        self.managers.append(manager)

    async def measure(self, method: BaseTechnique) -> list[Measurement]:
        """Concurrently start measurement on all managers in the pool.

        For hardware synchronization, set `use_hardware_sync` on the method.
        In addition, the pool must contain:
        - channels from a single multi-channel instrument only
        - the first channel of the multi-channel instrument
        - at least two channels

        All instruments are prepared and put in a waiting state.
        The measurements are started via a hardware sync trigger on channel 1.

        Parameters
        ----------
        method : MethodSettings
            Method parameters for measurement.
        """
        tasks: list[Awaitable[Measurement]] = []

        if hasattr(method, 'general') and method.general.use_hardware_sync:
            tasks = await self._measure_hw_sync(method)
        else:
            for manager in self.managers:
                tasks.append(manager.measure(method))

        results = await asyncio.gather(*tasks)
        return results

    async def _measure_hw_sync(
        self,
        method: BaseTechnique,
    ) -> list[Awaitable[Measurement]]:
        """Concurrently start measurement on all managers in the pool.

        Parameters
        ----------
        method : MethodSettings
            Method parameters for measurement.
        """
        follower_sync_tasks = []
        tasks = []

        for manager in self.managers:
            manager.validate_method(method._to_psmethod())

        if len(self.managers) < 2:
            raise ValueError(
                'Hardware synchronization requires two channels or more in the pool'
            )

        if len(set(manager.instrument.name for manager in self.managers)) > 1:
            raise ValueError(
                'Hardware synchronization is only supported when '
                'a single multi-channel instrument is selected.'
            )

        for manager in self.managers:
            if manager.instrument.channel == 1:
                hw_sync_manager = manager
                break
        else:
            raise ValueError(
                'Hardware synchronization requires the first channel '
                'of the multi-channel instrument to be in the pool.'
            )

        for manager in self.managers:
            if manager is hw_sync_manager:
                continue

            sync_task, measure_task = manager.initiate_hardware_sync_follower_channel(method)
            follower_sync_tasks.append(sync_task)
            tasks.append(measure_task)

        await asyncio.gather(*follower_sync_tasks)

        tasks.append(hw_sync_manager.measure(method))
        return tasks

    async def submit(self, func: Callable, **kwargs: Any) -> list[Any]:
        """Concurrently start measurement on all managers in the pool.

        Parameters
        ----------
        func : Callable
            This function gets called with an instance of
            `InstrumentManagerAsync` as the argument.
        **kwargs
            These keyword arguments are passed on to the submitted function.
        """
        tasks: list[Awaitable[Any]] = []
        for manager in self.managers:
            tasks.append(func(manager, **kwargs))

        results = await asyncio.gather(*tasks)
        return results



from __future__ import annotations

import os
from contextlib import contextmanager
from pathlib import Path
from typing import TYPE_CHECKING, Union

from PalmSens.Data import SessionManager
from PalmSens.DataFiles import MethodFile, MethodFile2
from System.IO import StreamReader, StreamWriter
from System.Text import Encoding

from ._data.measurement import Measurement
from ._methods import BaseTechnique, Method

if TYPE_CHECKING:
    from ._methods.method import Method


@contextmanager
def stream_reader(*args, **kwargs):
    sr = StreamReader(*args, **kwargs)
    try:
        yield sr
    finally:
        sr.Close()


@contextmanager
def stream_writer(*args, **kwargs):
    sw = StreamWriter(*args, **kwargs)
    try:
        yield sw
    finally:
        sw.Close()


def load_session_file(
    path: Union[str, Path],
) -> list[Measurement]:
    """Load a session file (.pssession).

    Parameters
    ----------
    path : Path | str
        Path to session file

    Returns
    -------
    measurements : list[Measurement]
        Return list of measurements
    """
    path = Path(path)

    session = SessionManager()

    with stream_reader(str(path)) as stream:
        session.Load(stream.BaseStream, str(path))

    session.MethodForEditor.MethodFilename = str(path.absolute())

    for psmeasurement in session:
        psmeasurement.Method.MethodFilename = str(path.absolute())

    return [Measurement(psmeasurement=m) for m in session]


def save_session_file(path: Union[str, Path], measurements: list[Measurement]):
    """Load a session file (.pssession).

    Parameters
    ----------
    path : Path | str
        Path to save the session file
    measurements : list[Measurement]
        List of measurements to save
    """
    path = Path(path)

    if any((measurement is None) for measurement in measurements):
        raise ValueError('cannot save null measurement')

    session = SessionManager()
    session.MethodForEditor = measurements[0]._psmeasurement.Method
    session.MethodForEditor.MethodFilename = str(path.absolute())

    for measurement in measurements:
        session.AddMeasurement(measurement._psmeasurement)

    with stream_writer(str(path), False, Encoding.Unicode) as stream:
        session.Save(stream.BaseStream, str(path))


def load_method_file(path: Union[str, Path], as_method: bool = False) -> BaseTechnique | Method:
    """Load a method file (.psmethod).

    Parameters
    ----------
    path : Path | str
        Path to method file
    as_method : bool
        If True, load as method wrapper object

    Returns
    -------
    method : Parameters
        Return method parameters
    """
    path = Path(path)

    with stream_reader(str(path)) as stream:
        if path.suffix == MethodFile2.FileExtension:
            psmethod = MethodFile2.FromStream(stream)
        else:
            psmethod = MethodFile.FromStream(stream, str(path))

    psmethod.MethodFilename = str(path.absolute())

    method = Method(psmethod=psmethod)

    if as_method:
        return method
    else:
        return method.to_settings()


def save_method_file(path: Union[str, Path], method: Union[Method, BaseTechnique]):
    """Load a method file (.psmethod).

    Parameters
    ----------
    path : Path | str
        Path to save the method file
    method : Method
        Method to save
    """
    from pypalmsens import __sdk_version__

    if isinstance(method, BaseTechnique):
        psmethod = method._to_psmethod()
    elif isinstance(method, Method):
        psmethod = method.psmethod
    else:
        raise ValueError(f'Unknown data type: {type(method)}')

    path = Path(path)

    with stream_writer(str(path), False, Encoding.Unicode) as stream:
        MethodFile2.Save(psmethod, stream.BaseStream, str(path), True, __sdk_version__)


def read_notes(path: Union[str, Path], n_chars: int = 3000):
    with open(path, encoding='utf16') as myfile:
        contents = myfile.read()
    raw_txt = contents[1:n_chars].split('\\r\\n')
    notes_list = [x for x in raw_txt if 'NOTES=' in x]
    notes_txt = (
        notes_list[0].replace('%20', ' ').replace('NOTES=', '').replace('%crlf', os.linesep)
    )
    return notes_txt



from __future__ import annotations

from abc import abstractmethod
from typing import ClassVar, Protocol, Type, runtime_checkable

import attrs
from PalmSens import Method as PSMethod
from PalmSens.Techniques import MixedMode as PSMixedMode

from pypalmsens._shared import single_to_double

from . import mixins
from ._shared import (
    CURRENT_RANGE,
)
from .base import BaseTechnique


@runtime_checkable
class BaseStage(Protocol):
    """Protocol to provide base methods for stage classes."""

    __attrs_attrs__: ClassVar[list[attrs.Attribute]] = []
    _type: int
    _registry: dict[int, Type[BaseStage]] = {}

    def __init_subclass__(cls, **kwargs) -> None:
        super().__init_subclass__(**kwargs)
        cls._registry[cls._type] = cls

    @classmethod
    def from_stage_type(cls, id: int) -> BaseStage:
        """Create new instance of appropriate stage from its type."""
        new = cls._registry[id]
        return new()

    @classmethod
    def _from_psstage(cls, psstage: PSMethod, /) -> BaseStage:
        """Generate parameters from dotnet method object."""
        new = cls.from_stage_type(psstage.StageType)
        new._update_params(psstage)
        new._update_params_nested(psstage)
        return new

    @abstractmethod
    def _update_params(self, psstage: PSMethod, /) -> None: ...

    def _update_params_nested(self, psstage, /) -> None:
        """Retrieve and convert dotnet method for nested field parameters."""
        for field in self.__attrs_attrs__:
            attribute = getattr(self, field.name)
            try:
                # Update parameters if attribute has the `update_params` method
                attribute._update_params(psstage)
            except AttributeError:
                pass

    def _update_psmethod(self, psmethod, /) -> PSMethod:
        """Add stage to dotnet method, and update paramaters on dotnet stage."""
        psstage = psmethod.AddStage(self._type)
        self._update_psstage(psstage)
        self._update_psstage_nested(psstage)
        return psstage

    @abstractmethod
    def _update_psstage(self, psstage: PSMethod, /) -> None: ...

    def _update_psstage_nested(self, psstage, /) -> None:
        """Convert and set field parameters on dotnet method."""
        for field in self.__attrs_attrs__:
            attribute = getattr(self, field.name)
            try:
                # Update parameters if attribute has the `update_params` method
                attribute._update_psmethod(psstage)
            except AttributeError:
                pass


@attrs.define(slots=False)
class ConstantE(BaseStage, mixins.CurrentLimitsMixin):
    """Amperometric detection stage."""

    _type = PSMixedMode.EnumMixedModeStageType.ConstantE

    potential: float = 0.0
    """Potential in V."""

    run_time: float = 1.0
    """Run time in s."""

    def _update_psstage(self, psstage, /):
        psstage.Potential = self.potential
        psstage.RunTime = self.run_time

    def _update_params(self, psstage, /):
        self.potential = single_to_double(psstage.Potential)
        self.run_time = single_to_double(psstage.RunTime)


@attrs.define(slots=False)
class ConstantI(BaseStage, mixins.PotentialLimitsMixin):
    """Potentiometry stage."""

    _type = PSMixedMode.EnumMixedModeStageType.ConstantI

    current: float = 0.0
    """The current to apply in the given current range.

    Note that this value acts as a multiplier in the applied current range.

    So if 10 uA is the applied current range and 1.5 is given as current value,
    the applied current will be 15 uA."""

    applied_current_range: CURRENT_RANGE = CURRENT_RANGE.cr_100_uA
    """Applied current range.

    Use `CURRENT_RANGE` to define the range."""

    run_time: float = 1.0
    """Run time in s."""

    def _update_psstage(self, psstage, /):
        psstage.AppliedCurrentRange = self.applied_current_range._to_psobj()
        psstage.Current = self.current
        psstage.RunTime = self.run_time

    def _update_params(self, psstage, /):
        self.applied_current_range = CURRENT_RANGE._from_psobj(psstage.AppliedCurrentRange)
        self.current = single_to_double(psstage.Current)
        self.run_time = single_to_double(psstage.RunTime)


@attrs.define(slots=False)
class SweepE(BaseStage, mixins.CurrentLimitsMixin):
    """Linear sweep detection stage."""

    _type = PSMixedMode.EnumMixedModeStageType.SweepE

    begin_potential: float = -0.5
    """Begin potential in V."""

    end_potential: float = 0.5
    """End potential in V."""

    step_potential: float = 0.1
    """Step potential in V."""

    scanrate: float = 1.0
    """Scan rate in V/s."""

    def _update_psstage(self, psstage, /):
        psstage.BeginPotential = self.begin_potential
        psstage.EndPotential = self.end_potential
        psstage.StepPotential = self.step_potential
        psstage.Scanrate = self.scanrate

    def _update_params(self, psstage, /):
        self.begin_potential = single_to_double(psstage.BeginPotential)
        self.end_potential = single_to_double(psstage.EndPotential)
        self.step_potential = single_to_double(psstage.StepPotential)
        self.scanrate = single_to_double(psstage.Scanrate)


@attrs.define(slots=False)
class OpenCircuit(BaseStage, mixins.PotentialLimitsMixin):
    """Ocp stage."""

    _type = PSMixedMode.EnumMixedModeStageType.OpenCircuit

    run_time: float = 1.0
    """Run time in s."""

    def _update_psstage(self, psstage, /):
        psstage.RunTime = self.run_time

    def _update_params(self, psstage, /):
        self.run_time = single_to_double(psstage.RunTime)


@attrs.define(slots=False)
class Impedance(BaseStage):
    """Electostatic impedance stage."""

    _type = PSMixedMode.EnumMixedModeStageType.Impedance

    run_time: float = 10.0
    """Run time in s."""

    dc_potential: float = 0.0
    """DC potential in V."""

    ac_potential: float = 0.01
    """AC potential in V RMS."""

    frequency: float = 50000.0
    """Frequency in Hz."""

    min_sampling_time: float = 0.5
    """Minimum sampling time in s.

    The instrument will measure at leas 2 sine waves.
    The sampling time will be automatically adjusted when necessary."""

    max_equilibration_time: float = 5.0
    """Max equilibration time in s.

    Used as a guard when the frequency drops below 1/max. equilibration time."""

    def _update_psstage(self, psstage, /):
        psstage.Potential = self.dc_potential
        psstage.Eac = self.ac_potential

        psstage.RunTime = self.run_time
        psstage.FixedFrequency = self.frequency

        psstage.SamplingTime = self.min_sampling_time
        psstage.MaxEqTime = self.max_equilibration_time

    def _update_params(self, psstage, /):
        self.dc_potential = single_to_double(psstage.Potential)
        self.ac_potential = single_to_double(psstage.Eac)

        self.run_time = single_to_double(psstage.RunTime)
        self.frequency = single_to_double(psstage.FixedFrequency)

        self.min_sampling_time = single_to_double(psstage.SamplingTime)
        self.max_equilibration_time = single_to_double(psstage.MaxEqTime)


@attrs.define
class MixedMode(
    BaseTechnique,
    mixins.CurrentRangeMixin,
    mixins.PretreatmentMixin,
    mixins.PostMeasurementMixin,
    mixins.DataProcessingMixin,
    mixins.GeneralMixin,
):
    """Create mixed mode method parameters."""

    _id = 'mm'

    interval_time: float = 0.1
    """Interval time in s."""

    cycles: int = 1
    """Number of times to go through all stages."""

    stages: list[BaseStage] = attrs.field(factory=list)
    """List of stages to run through."""

    def _update_psmethod(self, psmethod: PSMethod, /):
        """Update method with mixed mode settings."""
        psmethod.nCycles = self.cycles
        psmethod.IntervalTime = self.interval_time

        for stage in self.stages:
            stage._update_psmethod(psmethod)

    def _update_params(self, psmethod: PSMethod, /):
        self.cycles = psmethod.nCycles
        self.interval_time = single_to_double(psmethod.IntervalTime)

        for psstage in psmethod.Stages:
            stage = BaseStage._from_psstage(psstage)

            self.stages.append(stage)



from __future__ import annotations

import numpy as np


def single_to_double(val: float) -> float:
    """Cast single precision to double precision.

    Pythonnet returns System.Single, whereas python defaults to double precision.
    This leads to incorrect rounding, which makes comparing values difficult.

    By going through np.float32/str we can correctly round back to python float
    (which is double precision)."""
    return float(str(np.float32(val)))



from __future__ import annotations

import attrs

from . import settings
from ._shared import CURRENT_RANGE, POTENTIAL_RANGE


def current_converter(value: CURRENT_RANGE | settings.CurrentRange) -> settings.CurrentRange:
    if isinstance(value, CURRENT_RANGE):
        return settings.CurrentRange(min=value, max=value, start=value)
    return value


@attrs.define(slots=False)
class CurrentRangeMixin:
    current_range: settings.CurrentRange = attrs.field(
        factory=settings.CurrentRange, converter=current_converter
    )
    """Set the autoranging current."""


def potential_converter(
    value: POTENTIAL_RANGE | settings.PotentialRange,
) -> settings.PotentialRange:
    if isinstance(value, POTENTIAL_RANGE):
        return settings.PotentialRange(min=value, max=value, start=value)
    return value


@attrs.define(slots=False)
class PotentialRangeMixin:
    potential_range: settings.PotentialRange = attrs.field(
        factory=settings.PotentialRange, converter=potential_converter
    )
    """Set the autoranging potential."""


@attrs.define(slots=False)
class PretreatmentMixin:
    pretreatment: settings.Pretreatment = attrs.field(factory=settings.Pretreatment)
    """Set the pretreatment settings."""


@attrs.define(slots=False)
class VersusOCPMixin:
    versus_ocp: settings.VersusOCP = attrs.field(factory=settings.VersusOCP)
    """Set the versus OCP settings."""


@attrs.define(slots=False)
class BiPotMixin:
    bipot: settings.BiPot = attrs.field(factory=settings.BiPot)
    """Set the bipot settings"""


@attrs.define(slots=False)
class PostMeasurementMixin:
    post_measurement: settings.PostMeasurement = attrs.field(factory=settings.PostMeasurement)
    """Set the post measurement settings."""


@attrs.define(slots=False)
class CurrentLimitsMixin:
    current_limits: settings.CurrentLimits = attrs.field(factory=settings.CurrentLimits)
    """Set the current limit settings."""


@attrs.define(slots=False)
class PotentialLimitsMixin:
    potential_limits: settings.PotentialLimits = attrs.field(factory=settings.PotentialLimits)
    """Set the potential limit settings"""


@attrs.define(slots=False)
class ChargeLimitsMixin:
    charge_limits: settings.ChargeLimits = attrs.field(factory=settings.ChargeLimits)
    """Set the charge limit settings"""


@attrs.define(slots=False)
class IrDropCompensationMixin:
    ir_drop_compensation: settings.IrDropCompensation = attrs.field(
        factory=settings.IrDropCompensation
    )
    """Set the iR drop compensation settings."""


@attrs.define(slots=False)
class EquilibrationTriggersMixin:
    equilibrion_triggers: settings.EquilibrationTriggers = attrs.field(
        factory=settings.EquilibrationTriggers
    )
    """Set the trigger at equilibration settings."""


@attrs.define(slots=False)
class MeasurementTriggersMixin:
    measurement_triggers: settings.MeasurementTriggers = attrs.field(
        factory=settings.MeasurementTriggers
    )
    """Set the trigger at measurement settings."""


@attrs.define(slots=False)
class DelayTriggersMixin:
    delay_triggers: settings.DelayTriggers = attrs.field(factory=settings.DelayTriggers)
    """Set the delayed trigger at measurement settings."""


@attrs.define(slots=False)
class MultiplexerMixin:
    multiplexer: settings.Multiplexer = attrs.field(factory=settings.Multiplexer)
    """Set the multiplexer settings"""


@attrs.define(slots=False)
class DataProcessingMixin:
    data_processing: settings.DataProcessing = attrs.field(factory=settings.DataProcessing)
    """Set the data processing settings."""


@attrs.define(slots=False)
class GeneralMixin:
    general: settings.General = attrs.field(factory=settings.General)
    """Sets general/other settings."""
